# -*- coding: utf-8 -*-

"""Additional types used by NucleusAPI."""

from __future__ import annotations

__all__ = ()

import typing

if typing.TYPE_CHECKING:
    import typing_extensions

    class NucleusEnvironmentRecord(typing_extensions.TypedDict):
        """Description of the environment reported by Nucleus API."""

        id: str  # pylint: disable=invalid-name
        name: str
        yaml_ref: str
        revision: int
        created_at: str
        updated_at: str

    class NucleusEnvironmentCollection(typing_extensions.TypedDict):
        """Collection of environments reported by Nucleus API."""

        items: typing.Sequence[NucleusEnvironmentRecord]
        total: int
